from typing import Union, List, TypeVar

import torch
from torch.nn.utils.rnn import pad_sequence
from torch.utils.data.dataloader import DataLoader, Dataset

from transformers_rlfh.pad import left_pad_sequence
from transformers_rlfh.types import PPOSample, PPOBatch

__all__ = ['SampleStore']


def ppo_collator(samples: List[PPOSample], pad_token: int, **kwargs) -> PPOBatch:
    return PPOBatch(
        query=left_pad_sequence([sample.query for sample in samples], padding_value=pad_token),
        response=pad_sequence([sample.response for sample in samples], batch_first=True, padding_value=pad_token),
        logits=pad_sequence([sample.logits for sample in samples], batch_first=True, padding_value=0.0),
        values=pad_sequence([sample.values for sample in samples], batch_first=True, padding_value=0.0),
        rewards=pad_sequence([sample.reward for sample in samples], batch_first=True, padding_value=0.0),
        pad_token=pad_token,
    )


g_collators = {
    PPOSample: ppo_collator,
}


class CollatorWithArgs:
    def __init__(self, collator, **kwargs):
        self._collator = collator
        self._kwargs = kwargs

    def __call__(self, samples):
        return self._collator(samples, **self._kwargs)


T = TypeVar('T', covariant=True, bound=Union[PPOSample])


class SampleStore(Dataset[T]):
    def __init__(self, device=torch.device("cpu")):
        self._data = []
        self._device = device
        self._type = None

    def append(self, sample: T):
        if self._type is None:
            self._type = type(sample)
        else:
            if not isinstance(sample, self._type):
                raise TypeError(f"Expected {self._type}, got {type(sample)}")
        self._data.append(sample.to(self._device))

    def __getitem__(self, index) -> T:
        return self._data[index]

    def __len__(self) -> int:
        return len(self._data)

    def create_data_loader(self,
                           batch_size: int,
                           shuffle: bool = True,
                           pin_memory: bool = torch.cuda.is_available(), **kwargs) -> DataLoader:
        collator = g_collators[self._type]
        return DataLoader(self, batch_size=batch_size, shuffle=shuffle, pin_memory=pin_memory,
                          collate_fn=CollatorWithArgs(collator, **kwargs))



from typing import Callable, Iterator, List, Optional, Union, Iterable, TypeVar, Dict, Generator
from transformers_rlfh.store import SampleStore
from transformers_rlfh.types import PPOSample, ActorCriticOutput
from transformers import GenerationMixin, PreTrainedTokenizer, PreTrainedTokenizerFast
from transformers.utils import ModelOutput
import itertools
import torch
import dataclasses
from torchtyping import TensorType
import contextlib
from transformers_rlfh.pad import left_pad_sequence
from torch.nn.utils.rnn import pad_sequence

__all__ = ['generate_ppo_samples']

Tokenizer = Union[PreTrainedTokenizerFast, PreTrainedTokenizer]
T = TypeVar('T')


def chunk_n(iterable: Iterable[T], n: int) -> Iterator[List[T]]:
    """
    chunk_n 把一个iterable转换成 iterable of batched samples。n为最大的batch size
    :param iterable:
    :param n:
    :return:
    """
    it = iter(iterable)
    while True:
        chunk = list(itertools.islice(it, n))
        if not chunk:
            return
        yield chunk


@contextlib.contextmanager
def gpt_pad_left_special_tokenizer(tokenizer: Tokenizer):
    """
    GPT left padding的tokenizer的context manager。
    """
    if tokenizer.verbose:
        tokenizer.verbose = False
        verbose_disabled = True
    else:
        verbose_disabled = False
    # save the original pad token，if original pad token is None, we also disable verbose to supress warning
    pre_pad_token = tokenizer.pad_token
    if verbose_disabled:
        tokenizer.verbose = True
    tokenizer.pad_token = tokenizer.eos_token
    tokenizer.padding_side = "left"
    yield tokenizer
    tokenizer.pad_token = pre_pad_token
    tokenizer.padding_side = "right"


@dataclasses.dataclass
class GenerateSample:
    """
    Generation模型生成的样本
    """
    query: str
    response: str
    query_ids: TensorType["query_length", torch.long]
    response_ids: TensorType["response_length", torch.long]


def _generate_response_by_generation_model(
        query: Iterator[str],
        tokenizer: Tokenizer,
        generation_model: GenerationMixin,
        generate_batch_size: int,
        max_query_length: int,
        max_response_length: int,
        generate_kwargs: dict,
) -> Generator[GenerateSample, None, None]:
    """
    用generation model生成response
    """
    for query_text in chunk_n(query, generate_batch_size):
        generate_device = generation_model.device
        cpu = torch.device("cpu")
        with gpt_pad_left_special_tokenizer(tokenizer) as pad_left_tokenizer:
            cpu_query_batch = pad_left_tokenizer(query_text, padding=True, add_special_tokens=False,
                                                 max_length=max_query_length, return_tensors='pt')
        input_ids = cpu_query_batch["input_ids"].to(generate_device)
        attention_mask = cpu_query_batch["attention_mask"].to(generate_device)
        eos_id = tokenizer.eos_token_id
        with torch.no_grad():
            generation_output: ModelOutput = generation_model.generate(
                input_ids=input_ids,
                # max_length is the total length of the query and response
                max_length=max_response_length + input_ids.shape[1],
                attention_mask=attention_mask,
                return_dict_in_generate=True,
                **generate_kwargs,
            )
            sequences_ids = generation_output["sequences"][:, input_ids.shape[1]:].to(cpu)
            num_return_sequences: int = generate_kwargs['num_return_sequences']
            sequences = tokenizer.batch_decode(sequences_ids, skip_special_tokens=True)

            for i, query in enumerate(query_text):
                for j in range(num_return_sequences):
                    # un-padding query_ids and response_ids
                    response_ids = sequences_ids[i * num_return_sequences + j]
                    t = (response_ids == eos_id).nonzero().tolist()
                    if len(t) != 0:
                        response_ids = response_ids[:t[0][0]]

                    query_ids = cpu_query_batch['input_ids'][i]
                    t = (query_ids == eos_id).nonzero().tolist()
                    if len(t) != 0:
                        query_ids = query_ids[t[-1][0] + 1:]

                    yield GenerateSample(
                        query=query,
                        response=sequences[i * num_return_sequences + j],
                        query_ids=query_ids,
                        response_ids=response_ids,
                    )


@dataclasses.dataclass
class ScoredSample:
    """
    评分后的样本
    """
    sample: GenerateSample
    score: float


def _score_generated_sample(sample: Iterator[GenerateSample],
                            scorer: Callable[[List[str], List[str]], List[float]],
                            scorer_batch_size: int) -> Generator[ScoredSample, None, None]:
    """
    评分生成的样本
    """
    for batch in chunk_n(sample, scorer_batch_size):
        scores = scorer([sample.query for sample in batch], [sample.response for sample in batch])
        for sample, score in zip(batch, scores):
            yield ScoredSample(sample, score)


def _call_ref_model(scored_sample: Iterator[ScoredSample],
                    reference_model: Callable[..., ActorCriticOutput],
                    reference_batch_size: int,
                    eos_token_id: int) -> Generator[PPOSample, None, None]:
    """
    用reference model对样本进行评分
    """

    for batch in chunk_n(scored_sample, reference_batch_size):
        query_ids = left_pad_sequence([sample.sample.query_ids for sample in batch], padding_value=eos_token_id)
        response_ids = pad_sequence([sample.sample.response_ids for sample in batch], padding_value=eos_token_id,
                                    batch_first=True)
        max_query_len = query_ids.shape[1]
        input_ids = torch.cat([query_ids, response_ids], dim=1)
        attention_mask = input_ids.not_equal(eos_token_id).long()
        device = reference_model.device
        input_ids = input_ids.to(device)
        attention_mask = attention_mask.to(device)
        with torch.no_grad():
            reference_model.eval()
            output = reference_model(input_ids=input_ids, attention_mask=attention_mask)
            output.logits = output.logits.to("cpu")
            output.values = output.values.to("cpu")

            for i, sample in enumerate(batch):
                reward = torch.zeros(size=sample.sample.response_ids.shape, dtype=torch.float)
                reward[-1] = sample.score

                yield PPOSample(
                    query=sample.sample.query_ids,
                    response=sample.sample.response_ids,
                    logits=output.logits[i][max_query_len: max_query_len + len(sample.sample.response_ids)],
                    values=output.values[i][max_query_len: max_query_len + len(sample.sample.response_ids)],
                    reward=reward,
                )


def _adjust_reward(ppo_samples: Iterator[PPOSample],
                   adjust_reward: Optional[Callable[[List[PPOSample]], None]],
                   adjust_reward_batch_size: int) -> Generator[PPOSample, None, None]:
    """
    调整reward
    """
    if adjust_reward is None:
        for sample in ppo_samples:
            yield sample
    else:
        for batch in chunk_n(ppo_samples, adjust_reward_batch_size):
            adjust_reward(batch)
            for sample in batch:
                yield sample


def generate_ppo_samples(
        query: Iterator[str],
        tokenizer: Tokenizer,
        scorer: Callable[[List[str], List[str]], List[float]],
        generation_model: GenerationMixin,
        reference_model: Callable[..., ActorCriticOutput],
        max_query_length: int,
        max_response_length: int,
        adjust_reward: Optional[Callable[[List[PPOSample]], None]] = None,
        generate_batch_size: int = 8,
        generate_kwargs: Optional[Dict] = None,
        scorer_batch_size: int = 8,
        reference_batch_size: int = 8,
        adjust_reward_batch_size: int = 8,
) -> SampleStore[PPOSample]:
    """
    生成PPO samples for training

    :param query: query iterator
    :param tokenizer: 分词器
    :param scorer: 评价函数
    :param generation_model: 生成模型
    :param reference_model: 参考模型, 用于计算参考模型的 prob, value
    :param max_query_length: query最大长度
    :param max_response_length: response最大长度
    :param adjust_reward: 调整reward的函数, 用于增加 kl散度约束。可为空
    :param generate_batch_size:  生成过程的batch size
    :param generate_kwargs: 生成过程的参数
    :param scorer_batch_size: 评价过程的batch size
    :param reference_batch_size: 参考模型的batch size
    :param adjust_reward_batch_size: 调整reward的batch size
    :return: 生成的PPO样本
    """
    # 设置默认生成配置
    if generate_kwargs is None:
        generate_kwargs = dict()
    if 'num_beams' not in generate_kwargs:
        generate_kwargs['num_beams'] = 4
    if 'num_return_sequences' not in generate_kwargs:
        generate_kwargs['num_return_sequences'] = 4
    if 'do_sample' not in generate_kwargs:
        generate_kwargs['do_sample'] = False
    if 'early_stopping' not in generate_kwargs:
        generate_kwargs['early_stopping'] = True

    store = SampleStore[PPOSample]()

    # 生成PPOSample的过程主要分为4步
    # 1. 使用 generation model生成 response 文本
    # 2. 使用 scorer评价 query + response 文本
    # 3. 调用 reference model计算 prob, value
    # 4. 调整 reward， 例如增加KL散度约束等
    #
    # NOTE: 这四个步骤目前都可以设置独立的batch size，故使用generator做中间过程
    # NOTE: 每一步的输出尽量干净。即使某一步内部计算的时候需要padding，也应该在输出的时候去掉padding
    # NOTE: 因为可能生成很多PPOSample，所以中间结果都放到CPU中
    # NOTE: 如果速度不够，可以考虑多线程的生产者消费者模式

    generated_samples = _generate_response_by_generation_model(
        query=query, tokenizer=tokenizer, generation_model=generation_model,
        max_response_length=max_response_length, max_query_length=max_query_length,
        generate_batch_size=generate_batch_size, generate_kwargs=generate_kwargs)
    scored_samples = _score_generated_sample(generated_samples, scorer, scorer_batch_size)
    ppo_samples = _call_ref_model(scored_samples, reference_model, reference_batch_size, tokenizer.eos_token_id)
    for ppo_sample in _adjust_reward(ppo_samples, adjust_reward, adjust_reward_batch_size):
        store.append(ppo_sample)
    return store



from typing import Optional

import torch
from torchtyping import TensorType
from transformers import GPT2Model
from transformers.modeling_outputs import BaseModelOutputWithCrossAttentions

from transformers_rlfh.models.actor_critic import ValueHead

__all__ = ['GPTBestOfN']


class GPTBestOfN(torch.nn.Module):
    """
    GPT 排序模型，返回best of n

    参考 OpenAI LM-From-Human-preference
    """

    def __init__(self, base: GPT2Model, value_dropout=0.1):
        super().__init__()
        self.base = base
        self.value = ValueHead(feature_dim=base.config.hidden_size, dropout=value_dropout)
        self.value.to(self.base.device)
        self.reward_gain = torch.ones(size=(1,), dtype=torch.float32, device=self.base.device)
        self.reward_bias = torch.zeros(size=(1,), dtype=torch.float32, device=self.base.device)

    @property
    def device(self):
        return self.base.device

    def forward(self,
                input_ids: TensorType["batch_size", "n_best", "sequence_length", torch.long],
                best: TensorType["batch_size", torch.long],
                last_token_pos: Optional[TensorType["batch_size"]] = None,
                pad_token_id: Optional[int] = None,
                loss_reduction="mean",
                ):
        """

        :param input_ids: batch_size x n_best x query_length
        :param best: best sample id
        :param last_token_pos: batch_size, 每个sample最后一个token的id. 包含special token。只用来计算last token reward
        :param pad_token_id: pad token id, Padding token id
        :param loss_reduction: loss 的reduction方法。用于单元测试

        :note: last_token_ids, pad_token_id 要么全部设置，要么全部不设置。如果设置，会计算last token reward。
               否则直接计算最后一个padding的reward

        :return: loss
        """
        n_best = input_ids.shape[1]
        batch_size = input_ids.shape[0]
        sequence_length = input_ids.shape[2]
        reward: TensorType["batch_size*n_best"] = self.get_reward(
            input_ids=input_ids.reshape(batch_size * n_best, sequence_length),
            pad_token_id=pad_token_id,
            last_token_pos=None if last_token_pos is None else last_token_pos.reshape(batch_size * n_best),
        )
        reward: TensorType["batch_size", "n_best"] = reward.reshape(batch_size, n_best)

        return torch.nn.functional.nll_loss(input=torch.nn.functional.log_softmax(reward, dim=1), target=best,
                                            reduction=loss_reduction)

    def get_reward(self,
                   input_ids: TensorType["batch_size", "sequence_length", torch.long],
                   last_token_pos: Optional[TensorType["batch_size"]] = None,
                   pad_token_id: Optional[int] = None) -> TensorType["batch_size", torch.float32]:
        """

        :param input_ids: batch_size x sequence_length query和response已经拼好了
        :param last_token_pos: batch_size, 每个sample最后一个token的id. 包含special token。只用来计算last token reward
        :param pad_token_id: pad token id, Padding token id

        :note: last_token_ids, pad_token_id 要么全部设置，要么全部不设置。如果设置，会计算last token reward。
               否则直接计算最后一个padding的reward
        :return: reward
        """
        # check (last_token_pos, pad_token_id) are both set or both not set
        if self.reward_gain.device != input_ids.device:
            self.reward_gain = self.reward_gain.to(input_ids.device)
            self.reward_bias = self.reward_bias.to(input_ids.device)


        if last_token_pos is None and pad_token_id is None:
            reward_style = "last_padding"
        elif last_token_pos is not None and pad_token_id is not None:
            reward_style = "last_token"
        else:
            raise ValueError("last_token_ids, pad_token_id must be all set or all not set")

        if reward_style == "last_padding":
            att_mask = torch.ones_like(input_ids, device=input_ids.device, dtype=torch.bool)
        else:
            att_mask = torch.ne(input_ids, pad_token_id)

        output: BaseModelOutputWithCrossAttentions = self.base(input_ids=input_ids, attention_mask=att_mask)
        last_hidden_state: ["batch_size", "sequence_length"] = output.last_hidden_state.squeeze(-1)

        if reward_style == "last_padding":
            last_hidden_state = last_hidden_state[:, -1, :]
            reward: TensorType["batch_size"] = self.value(last_hidden_state)
        else:
            mask = torch.nn.functional.one_hot(last_token_pos,
                                               num_classes=last_hidden_state.shape[1]).unsqueeze(-1).float()
            last_hidden_state = last_hidden_state * mask
            last_hidden_state = torch.sum(last_hidden_state, dim=1)
            reward: TensorType["batch_size"] = self.value(last_hidden_state)

        reward = self.reward_gain * reward + self.reward_bias
        return reward



from typing import Union, List, Optional

import torch
from transformers import PreTrainedTokenizerFast, AutoTokenizer

__all__ = ['BestOfNCollator']

Tokenizer = Union[PreTrainedTokenizerFast, AutoTokenizer]


class BestOfNCollator:
    """
    训练GPT Best Of N排序的collator

    """

    def __init__(self, tokenizer: Tokenizer, special_token: str, n_best: int = 4, pad_token_id: Optional[int] = None):
        if pad_token_id is None:
            pad_id = tokenizer.convert_tokens_to_ids([special_token])[0]
            tokenizer.pad_token_id = pad_id
            tokenizer.pad_token = special_token
            self._return_last_token_pos = False
        else:
            tokenizer.pad_token = tokenizer.convert_ids_to_tokens([pad_token_id])[0]
            tokenizer.pad_token_id = pad_token_id
            self._return_last_token_pos = True
        self._tokenizer = tokenizer
        self._special_token = special_token
        self._n_best = n_best

    def __call__(self, batch):
        texts = ["".join([item["query"].strip(), self._special_token, item[f"sample{i}"].strip(), self._special_token])
                 for item in batch for i in range(self._n_best)]

        if not self._return_last_token_pos:
            texts = self._tokenizer(texts, padding=True, return_tensors="pt",
                                    return_attention_mask=False, return_token_type_ids=False)
            input_ids = texts["input_ids"]
        else:
            texts = self._tokenizer(texts, padding=False, return_token_type_ids=False,
                                    return_length=True)
            last_token_pos = texts["length"]
            last_token_pos = torch.tensor(last_token_pos) - 1
            last_token_pos.resize_(len(batch), self._n_best)
            input_ids = self._tokenizer.pad({"input_ids": texts["input_ids"]}, return_tensors="pt", padding=True,
                                            return_attention_mask=False)["input_ids"]

        input_ids.resize_(len(batch), self._n_best, input_ids.shape[1])
        returns = [input_ids, torch.tensor([elem["best"] for elem in batch], dtype=torch.long)]

        if self._return_last_token_pos:
            returns.append(last_token_pos)

        return returns



import torch
from torchtyping import TensorType
import dataclasses

__all__ = ['PPOSample', 'PPOBatch', 'ActorCriticOutput']


@dataclasses.dataclass
class PPOSample:
    """
    query: Query的tensor。query length每个sample可以不同
    response: Response的tensor。response length每个sample可以不同
    reward: 每一个token的reward。其实大多数情况下，只有sequence结尾的时候给一个score就行了。不过为了增加 kl散度的惩罚项。所以每一个token
            初始化成KL散度的惩罚项。
    values: actor-critic网络中，critic的输出。这个values是 ref model的values
    logits: actor-critic网络中，actor的输出。即不同token的概率。 这个probs是ref model的probs
    """
    query: TensorType["query_length", torch.long]
    response: TensorType["response_length", torch.long]
    reward: TensorType["response_length", torch.float]
    values: TensorType["response_length", torch.float]
    logits: TensorType["response_length", "vocab_size", torch.float]

    def to(self, device: torch.device) -> 'PPOSample':
        return PPOSample(
            query=self.query.to(device),
            response=self.response.to(device),
            reward=self.reward.to(device),
            values=self.values.to(device),
            logits=self.logits.to(device),
        )


@dataclasses.dataclass
class PPOBatchModelInput:
    """
    送给GPT模型的输入

    input_ids: 组合query和response的tensor
    attention_mask: attention的mask
    """
    input_ids: TensorType["batch_size", "query_length+response_size", torch.long]
    attention_mask: TensorType["batch_size", "query_length+response_size", torch.long]


@dataclasses.dataclass
class PPOBatch:
    """
    values: actor-critic网络中，critic的输出。这个values是 ref model的values
    logits: actor-critic网络中，actor的输出。即不同token的概率。 这个probs是ref model的probs
    """
    query: TensorType["batch_size", "query_length", torch.long]  # left padding
    response: TensorType["batch_size", "response_length", torch.long]  # right padding
    logits: TensorType["batch_size", "response_size", "vocab_size", torch.long]
    values: TensorType["batch_size", "response_size", torch.float]
    rewards: TensorType["batch_size", "response_size", torch.float]
    pad_token: int

    def to_model_input(self) -> PPOBatchModelInput:
        input_ids = torch.cat([self.query, self.response], dim=1)
        attention_mask = input_ids.not_equal(self.pad_token)
        return PPOBatchModelInput(input_ids=input_ids, attention_mask=attention_mask)

    def to(self, device: torch.device) -> 'PPOBatch':
        return PPOBatch(
            query=self.query.to(device),
            response=self.response.to(device),
            logits=self.logits.to(device),
            values=self.values.to(device),
            rewards=self.rewards.to(device),
        )

    def pin_memory(self) -> 'PPOBatch':
        return PPOBatch(
            query=self.query.pin_memory(),
            response=self.response.pin_memory(),
            logits=self.logits.pin_memory(),
            values=self.values.pin_memory(),
            rewards=self.rewards.pin_memory(),
        )


@dataclasses.dataclass
class ActorCriticOutput:
    logits: TensorType["batch_size", "sequence_length", "vocab_size"]
    values: TensorType["batch_size", "sequence_length"]



import os.path

import datasets
from transformers_rlfh.datasets.rank_to_pairwise import convert_rank_dataset_to_pairwise


def test_rank_to_pairwise():
    path = os.path.join(os.path.dirname(__file__), "..", "..", "examples", "sentimental", "data")
    ds = datasets.load_from_disk(path)
    prev_len = len(ds["train"])
    ds = convert_rank_dataset_to_pairwise(ds)
    new_len = len(ds["train"])
    assert new_len == 3 * prev_len



from datasets import DatasetDict

__all__ = ['convert_rank_dataset_to_pairwise']


def mapper(batch):
    query = batch["query"]
    samples = zip(*[batch[f"sample{i}"] for i in range(4)])
    best = batch["best"]

    return_query = [None] * (len(query) * 3)
    return_good_sample = [None] * (len(query) * 3)
    return_bad_sample = [None] * (len(query) * 3)

    offset = 0

    for _query, _best, _samples in zip(query, best, samples):
        for i, sample in enumerate(_samples):
            if i == best:
                continue

            return_query[offset] = _query
            return_good_sample[offset] = _samples[_best]
            return_bad_sample[offset] = sample

    return {
        "query": return_query,
        "good_sample": return_good_sample,
        "bad_sample": return_bad_sample,
    }


def convert_rank_dataset_to_pairwise(ds: DatasetDict) -> DatasetDict:
    """
    将RLFH rank的数据集转换成pairwise数据集
    输入columns
        * query
        * sample0
        * sample1
        * sample2
        * sample3
        * best: int
    输出:
        query, good_sample, bad_sample

    每一个输入，可以产生三个输出。即 best 的sample，比其他的sample要更好
    """
    return ds.map(mapper, batched=True, remove_columns=["sample0", "sample1", "sample2", "sample3", "best", "query"])



from typing import List, Optional

import torch

__all__ = ['left_pad_sequence']


def left_pad_sequence(sequence, padding_value: int):
    max_length = max(seq.shape[0] for seq in sequence)
    stacks: List[Optional[torch.Tensor]] = [None] * len(sequence)
    for i, seq in enumerate(sequence):
        pad_length = max_length - seq.shape[0]
        if pad_length == 0:
            stacks[i] = seq
        else:
            with torch.no_grad():
                pad = torch.nn.ConstantPad1d((pad_length, 0), padding_value)
                seq = pad(seq)
                stacks[i] = seq
    return torch.stack(stacks)



import torch.nn
from transformers import GPTNeoForCausalLM
from transformers.modeling_outputs import CausalLMOutputWithPast
from transformers_rlfh.types import ActorCriticOutput


class ValueHead(torch.nn.Module):
    def __init__(self, feature_dim: int, dropout: float = 0.1):
        super().__init__()

        layers = []
        if dropout > 0:
            layers.append(torch.nn.Dropout(p=dropout))
        layers.append(torch.nn.Linear(feature_dim, 1))

        self.head = torch.nn.Sequential(*layers)

    def forward(self, x):
        return self.head(x)


class ActorCriticLM(torch.nn.Module):
    def __init__(self, base: GPTNeoForCausalLM):
        super().__init__()
        self.base = base
        self.value_head = ValueHead(base.config.hidden_size)
        self.value_head.to(self.base.device)

    @property
    def device(self):
        return self.base.device

    def forward(self, input_ids, attention_mask):
        output: CausalLMOutputWithPast = self.base(input_ids=input_ids, attention_mask=attention_mask,
                                                   output_hidden_states=True)
        value = self.value_head(output.hidden_states[-1]).squeeze(-1)
        return ActorCriticOutput(logits=output.logits, values=value)



import os
from typing import List, Union

import datasets
import torch.optim.optimizer
import tqdm
from torch.utils.data import DataLoader
from transformers import GPTNeoForCausalLM, AutoTokenizer

from transformers_rlfh.models.gpt_best_of_n import GPTBestOfN
from transformers_rlfh.datasets.best_of_n_collator import BestOfNCollator

SPECIAL_TOKEN = "<|end of req rsp|>"


def test_two_loss_for_same_input_are_same():
    model = GPTNeoForCausalLM.from_pretrained("EleutherAI/gpt-neo-125M")
    model = model.transformer
    tokenizer = AutoTokenizer.from_pretrained("EleutherAI/gpt-neo-125M")
    tokenizer.add_tokens([SPECIAL_TOKEN])
    model.resize_token_embeddings(len(tokenizer))
    device = torch.device("cuda:0")
    model = GPTBestOfN(base=model, value_dropout=0)
    model.to(device)

    loss = model(input_ids=torch.tensor([[[1, 2, 3, 4], [5, 6, 7, 8]], [[1, 2, 3, 4], [5, 6, 7, 8]]], dtype=torch.long,
                                        device=device),
                 best=torch.tensor([0, 0], dtype=torch.long, device=device), loss_reduction="none")
    assert len(loss.shape) == 1
    assert loss.shape[0] == 2
    l1 = loss[0].item()
    l2 = loss[1].item()
    assert l1 == l2


def test_train_gpt_best_of_n():
    model = GPTNeoForCausalLM.from_pretrained("EleutherAI/gpt-neo-125M")
    model = model.transformer
    tokenizer = AutoTokenizer.from_pretrained("EleutherAI/gpt-neo-125M")
    tokenizer.add_tokens([SPECIAL_TOKEN])
    model.resize_token_embeddings(len(tokenizer))
    collator = BestOfNCollator(tokenizer, special_token=SPECIAL_TOKEN, pad_token_id=tokenizer.eos_token_id)
    path = os.path.join(os.path.dirname(__file__), "..", "..", "examples", "sentimental", "data")
    ds = datasets.load_from_disk(path)
    dataloader = DataLoader(dataset=ds["train"], collate_fn=collator, batch_size=2)

    model = GPTBestOfN(base=model)
    device = torch.device("cuda:0")
    model.to(device)
    optimizer = torch.optim.Adam(model.parameters())

    for _ in range(1):
        for i, (input_ids, best, last_token_pos) in enumerate(tqdm.tqdm(dataloader)):
            loss = model(input_ids=input_ids.to(device), best=best.to(device), last_token_pos=last_token_pos.to(device),
                         pad_token_id=tokenizer.eos_token_id)
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            print(i + 1, loss)
            if i + 1 == 50:
                break



import itertools

from transformers import AutoModelForCausalLM, AutoTokenizer
from transformers_rlfh.generate import generate_ppo_samples
from transformers_rlfh.models.actor_critic import ActorCriticLM


def test_generate():
    generate_model = AutoModelForCausalLM.from_pretrained("EleutherAI/gpt-neo-125M")
    generate_model.to("cuda:0")
    tokenizer = AutoTokenizer.from_pretrained("EleutherAI/gpt-neo-125M")
    ac = ActorCriticLM(generate_model)

    store = generate_ppo_samples(
        ["EleutherAI has",
         "今天天气不错"],
        tokenizer, scorer=lambda q, r: [len(r_) for r_ in r], generation_model=generate_model,
        reference_model=ac,
        max_query_length=100, max_response_length=32)
    print(len(store) != 0)



from setuptools import setup, find_packages

setup(
    name="transformers-rlfh",
    version="0.0.1.dev.2",
    packages=find_packages(exclude=["tests", "tests.*", "test"]),
    description="RLFH with transformers",
    long_description="RLFH with transformers",
    install_requires=[
        "transformers",
        "torch",
        "numpy",
        "wandb",
        "optuna",
        "datasets",
        "tqdm",
        "torchtyping",
    ]
)
